# ------------------------------------------------------------------------------
# Calculate trajectory and spin of ping pong ball hit by paddle and bouncing
# on table.
#
# Use a friction and energy loss model for ball bounces to predict spin.
# Model aerodynamic drag, magnus effect, spin drag.
#
class Physical_Parameters:
    def __init__(self):
        s = self

        s.paddle_friction = 3              # coef of friction of paddle
        s.paddle_damping = 0.3             # energy loss in paddle bounce

        s.ball_radius = 0.02               # 40 mm diameter ball, meters
        s.ball_mass = 0.0027               # mass of ball, 2.7 grams

        s.table_height = 30 * 0.0254       # table height, 30 inches
        s.table_length = 9 * 12 * 0.0254   # table length, 9 feet
        s.table_friction = 0.25            # coef of friction of table
        s.table_damping = 0.2              # energy loss in table bounce
        s.net_height = 6 * 0.0254          # net height, 6 inches

        # Aerodynamics
        s.drag_coefficient = 0.5
        s.lift_coefficient = 0.28          # magnus effect lift coefficient
        s.spin_drag_coefficient = 0.012    # spin drag w*v dependence
        s.spin_down_coefficient = 3.5      # spin drag, no motion, w**1.5 term
        s.air_density = 1.27               # density of air, kg/m^3
        s.air_viscosity = 1.8e-5           # dynamic visocosity of air, kg/(m*s)

        s.gravity = 9.8                    # accel of gravity, m/sec^2

physical_params = Physical_Parameters()

class Ball_State:

    def __init__(self, time = 0, position = (0,0), velocity = (0,0), spin = 0,
                 incoming_curve = None, curve_index = None):
        self.time = time                # seconds
        self.position = position        # meters from floor at net
        self.velocity = velocity        # meters / sec
        self.spin = spin                # radians / sec
        self.incoming_curve = incoming_curve
        if curve_index is None and not incoming_curve is None:
            curve_index = len(incoming_curve) - 1
        self.curve_index = curve_index  # Position on incoming curve.

    def set_state(self, bs):
        self.time = bs.time
        self.position = bs.position
        self.velocity = bs.velocity
        self.spin = bs.spin
        self.incoming_curve = bs.incoming_curve
        self.curve_index = bs.curve_index

# ------------------------------------------------------------------------------
# Calculate paddle strike, ball flight, table, net and floor bounces.
#
def hit(t, p, vb = (0,0), w = 0, pa = 0, vp = (1,0),
        side = 'right', serve = True, pp = physical_params,
        tstep = 0.001, maxstep = 10000, maxpath = 10000):

    # Compute ball speed, spin after paddle strike.
    v, w, f = strike(vb, w, vp, pa, pp.ball_radius,
                     pp.paddle_damping, pp.paddle_friction)

    return flight(t, p, v, w, side, serve, pp, tstep, maxstep, maxpath)

# ------------------------------------------------------------------------------
# Calculate ball flight, table, net and floor bounces.
#
def flight(t, p, v, w, side = 'right', serve = True, pp = physical_params,
           tstep = 0.001, maxstep = 10000, maxpath = 10000):

    curves = []
    from numpy import empty, double
    pa = empty((max(1,maxpath),6), double)
    oside = 'left' if side == 'right' else 'right'
    seq = [('table',side), ('table',oside),'stop'] if serve else [('table',oside),'stop']
    r, et, mut = pp.ball_radius, pp.table_damping, pp.table_friction
    for b,sh in enumerate(seq):
        # Compute flight until table or floor strike.
        s,n = arc(t, p, v, w, pp, tstep, maxstep, pa)
        if maxpath > 0:
            curves.append(pa[:n,:].copy())
        t, px, py, vx, vy, w = pa[n-1,:]
        p = (px,py)
        hs = 'left' if px <= 0 else 'right'
        if (s,hs) != sh:
            break
        if maxpath <= 0 and b == len(seq)-2:
            sh = 'stop'         # Skip calculating path after ball lands fair.
            break
        # Compute bounce off table.
        v, w, fc = bounce((vx,vy), w, r, et, mut)

    good = (sh == 'stop')
    if maxpath <= 0:
        return good
    return good, curves

# ------------------------------------------------------------------------------
# Calculate ball drop from given height for serve.
#
def drop(hd, sp, side = 'right', serve = True, pp = physical_params,
         tstep = 0.001, maxstep = 10000):

    # Compute ball drop speed, spin.
    t = 0
    x,hs = sp
    p = (x,hs+hd)
    v = (0,0)
    w = 0

    # Fake the table height to calculate ball path to strike height.
    tl, th, r = pp.table_length, pp.table_height, pp.ball_radius
    pp.table_length = 3*abs(x)
    pp.table_height = hs-r

    from numpy import empty, double
    path = empty((maxstep,6), double)
    s,n = arc(t, p, v, w, pp, tstep, maxstep, path)

    pp.table_length, pp.table_height = tl, th

    return path[:n,:]

# ------------------------------------------------------------------------------
# Compute flight path of ball until table strike (y=th "table height")
# within +/-0.5*tl (half table length), net strike, or floor strike.
#
def arc(t, p, v, w, pp, tstep, maxstep, path):

    # Call C++ routine for fast calculation of trajectory
    from . import fly
    r = fly.flight(t, p[0], p[1], v[0], v[1], w,
                   pp.ball_radius, pp.ball_mass,
                   pp.table_height, pp.table_length, pp.net_height,
                   pp.drag_coefficient, pp.lift_coefficient,
                   pp.spin_down_coefficient, pp.spin_drag_coefficient,
                   pp.air_density, pp.air_viscosity, pp.gravity,
                   tstep, maxstep, path)
    return r


# ------------------------------------------------------------------------------
# Compute ball bounce off of paddle.
#
# vp = paddle velocity
# pa = paddle angle, radians of normal from front of paddle to x axis.
#
def strike(vb, w, vp, pa, r = 0.02, e = 0, mu = 1):

    # Transform to frame where paddle is motionless.
    vb2 = (vb[0]-vp[0],vb[1]-vp[1])

    # Rotate to frame where paddle is parallel to x axis.
    from math import pi, sin, cos
    a = -(pa - 0.5*pi)
    s,c = sin(a), cos(a)
    vb3 = (c*vb2[0] - s*vb2[1], s*vb2[0] + c*vb2[1])    # Rotate by a.

    vbf3, wf, fc = bounce(vb3, w, r, e, mu)
    vbf2 = (c*vbf3[0] + s*vbf3[1], -s*vbf3[0] + c*vbf3[1])    # Rotate by -a.
    vbf = (vbf2[0]+vp[0],vbf2[1]+vp[1])

    return vbf, wf, fc

# ------------------------------------------------------------------------------
# Bounce off y = 0 plane from above.
#
# vb = initial ball velocity
# w = initial rotation rate (radians / sec)
# r = ball radius
# e = fractional kinetic energy loss in bounce perpendicular to surface.
# mu = coeficient of sliding friction.
#
# Ball moment of inertia I = 2/3*m*R*R.
#
# Returns
#
# vout = final ball velocity
# wout = final rotational rate
# fc = friction needed to avoid skidding
#
def bounce(vb, w, r = 0.02, e = 0, mu = 1):

    vx, vy = vb
    if vy >= 0:
        return vb, w, 0       # No contact, ball moving away from surface.

    from math import sqrt
    f = sqrt(1 - e)

    uy = -f*vy

    vcx = vx + w*r    # tangential contact speed
    if abs(vcx) <= 2.5*mu*(1+f)*abs(vy):
        # Within friction limit, sticky contact.
        ux = vx - 0.4*vcx
        wout = w - 0.6*vcx/r
    else:
        # Exceeds friction limit, skidding contact.
        sign_vcx = (-1 if vcx < 0 else 1)
        pom = mu*(1+f)*abs(vy) * sign_vcx
        dvcx = 2.5*pom
        ux = vx - 0.4*dvcx
        wout = w - 0.6*dvcx/r

    fc = abs(vcx) / (2.5*(1+f)*abs(vy)) if vy != 0 else 1000

    return (ux,uy), wout, fc

# ------------------------------------------------------------------------------
# Find paddle angles and velocities that make the ball skim the net and touch
# the baseline.
#
def velocity_scan(p, vb = (0,0), w = 0, pa = 0, side = 'right', serve = True,
                  pp = physical_params, vrange = ((-10,10,1),(-10,10,1)),
                  tstep = 0.001):


    (vxmin, vxmax, vxstep), (vymin, vymax, vystep) = vrange

    vxsteps = 1 + int((vxmax-vxmin)/vxstep)
    vysteps = 1 + int((vymax-vymin)/vystep)

    from numpy import empty, int32, transpose, float32
    g = empty((vysteps, vxsteps), int32)
    IN, SHORT, LONG = 1, 2, 3

    t = 0
    maxpath = 1
    tlh = 0.5*pp.table_length
    clenlong = 2 if serve else 1
    for vxi in range(vxsteps):
        vx = vxmin + vxi*vxstep
        for vyi in range(vysteps):
            vy = vymin + vyi*vystep
            good, curves = hit(t, p, vb, w, pa, (vx,vy), side, serve, pp,
                               tstep, maxpath = maxpath)
            if good:
                r = IN
            else:
                x = curves[-1][-1][1]
                if len(curves) != clenlong:
                    r = SHORT
                elif ((side == 'right' and x < -tlh) or
                      (side == 'left' and x > tlh)):
                    r = LONG
                else:
                    r = SHORT
            g[vyi,vxi] = r
    return g

# ------------------------------------------------------------------------------
#
def drag_force(v, physical_params):

    p = physical_params
    r = p.ball_radius
    from math import pi
    A = pi*r*r
    v2 = v[0]*v[0] + v[1]*v[1]
    f = p.drag_coefficient * 0.5 * p.air_density * A * v2
    return f

# ------------------------------------------------------------------------------
#
def lift_force(v, w, physical_params):

    p = physical_params
    from math import sqrt, pi
    vmag = sqrt(v[0]*v[0] + v[1]*v[1])
    r = p.ball_radius
    A = pi*r*r
    f = p.lift_coefficient * 0.5 * p.air_density * A * w*r * vmag
    return f

# ------------------------------------------------------------------------------
# Compute paddle angle and velocity to produce a given ball velocity and spin.
#
def calculate_strike(vin, win, vout, wout,
                     paddle_friction, paddle_damping, ball_radius):

    v = [vo-vi for vo,vi in zip(vout,vin)]
    from math import sqrt, asin, sin, cos, atan2
    vn = sqrt(v[0]*v[0] + v[1]*v[1])
    r = ball_radius
    I = (2./3)*r*r
    dL = I*(wout-win)
    f = paddle_friction
    c = f / sqrt(f*f + 1)       # cosine non-skid angle
    if abs(dL) > c*vn*r:
        raise ValueError("Unachievable spin")
    a = -asin(dL/(vn*r))  # angle from paddle normal to ball velocity direction
    vpar = vn*sin(a) - wout*r
    e = paddle_damping
    vperp = vn*cos(a)/(sqrt(1-e) + 1)
    pa = atan2(v[1],v[0]) - a   # Angle from x-axis to paddle normal.
    vp = (vperp*cos(pa) - vpar*sin(pa) + vin[0],
          vperp*sin(pa) + vpar*cos(pa) + vin[1])
    return pa, vp

# ------------------------------------------------------------------------------
# Find paddle angle and velocity to achieve given height over net, depth of
# bounce and spin after bounce.  Use an iterative walk method starting from a
# current hit that is in play.
#
# TODO: convergence is poor.  Improve algorithm.
#
def hit_target(height_at_net, depth_of_bounce, post_bounce_spin,
               t, p, vb, w, pa, vp, side = 'right', serve = True,
               pp = physical_params,
               tstep = 0.001, maxstep = 10000, maxpath = 10000,
               htol = 0.005, dtol = 0.025, stol = 1.0):

    good, curves = hit(t, p, vb, w, pa, vp, side, serve,
                       pp, tstep, maxstep, maxpath)
    if not good:
        return pa, vp
    han, dob, pbs = receive_parameters(curves, serve, side)
    dh = abs(han - height_at_net)
    dd = abs(dob - depth_of_bounce)
    ds = abs(pbs - post_bounce_spin)

    vstep = 0.1         # m/s
    astep = 0.01        # radians
    i = cuts = 0
    max_tries = 1000
    max_cuts = 10
    from random import uniform
#    s = b = l = 0
#    ll = []
    while i < max_tries and cuts <= max_cuts:
        i += 1
        pa2 = pa + uniform(-astep, astep)
        vp2 = (vp[0] + uniform(-vstep, vstep), vp[1] + uniform(-vstep, vstep))
        good, curves = hit(t, p, vb, w, pa2, vp2, side, serve,
                           pp, tstep, maxstep, maxpath)
        if not good:
#            b += 1
            continue
        han2, dob2, pbs2 = receive_parameters(curves, serve, side)
        dh2 = abs(han2 - height_at_net)
        dd2 = abs(dob2 - depth_of_bounce)
        ds2 = abs(pbs2 - post_bounce_spin)
#        ll.append((dh2,dd2,ds2))
#        if (dh2 < dh or dh2 < htol) and (dd2 < dd or dd2 < dtol):
#            l += 1
        if ((dh2 < dh or dh2 < htol) and
            (dd2 < dd or dd2 < dtol) and
            (ds2 < ds or ds2 < stol)):
            dh, dd, ds = dh2, dd2, ds2
            pa = pa2
            vp = vp2
            pbsb, hanb, dobb = pbs2, han2, dob2
#            s += 1
            if dh2 < htol and dd2 < dtol and ds2 < stol:
                f = 0.5
                vstep *= f
                astep *= f
                htol *= f
                dtol *= f
                stol *= f
                cuts += 1

#    print ('steps %d, place ok %d, out %d, step cut %d' % (s, l, b, cuts))
#    print ('change h %.3f, d %.3f, s %.3f (%.3f)' % (hanb - han, dobb - dob, pbsb - pbs, post_bounce_spin - pbs))
#    from numpy import mean
#    print ('mean error h %.3f, d %.3f, s %.3f' % tuple(mean(ll, axis = 0)))

    return pa, vp

# ------------------------------------------------------------------------------
#
def receive_parameters(curves, serve, side):

    c = 1 if serve else 0
    h = height_at_zero(curves[c][:,1:3]) if c < len(curves) else None
    if c+1 < len(curves):
        b = curves[c+1][0]
        d = b[1] if side == 'left' else -b[1]
        s = b[5]
    else:
        d = s = None
    return h, d, s

# ------------------------------------------------------------------------------
#
def height_at_zero(xy):

    x = xy[:,0]
    y = xy[:,1]
    signs = (x > 0) if x[0] > 0 else (x <= 0)
    i = signs.argmin() - 1
    f = x[i] / (x[i] - x[i+1])
    h = (1-f) * y[i] + f * y[i+1]
    return h

# ------------------------------------------------------------------------------
#
def set_ball_state_at_time(bstate, t0, path):

    c, i = curve_at_time(path, t0)
    t,px,py,vx,vy,w = c[i]
    bs = bstate
    bs.time, bs.position, bs.velocity, bs.spin = t,(px,py),(vx,vy),w
    bs.incoming_curve = c
    bs.curve_index = i

# ------------------------------------------------------------------------------
#
def path_time_point(curves, t0):

    c, i = curve_at_time(curves, t0)
    if c is None:
        return None
    return c[i]

# ------------------------------------------------------------------------------
#
def curve_at_time(curves, t0):

    for c in curves:
        if t0 <= c[-1][0]:
            for i,s in enumerate(c):
                if s[0] >= t0:
                    return c,i
    return None, None

# ------------------------------------------------------------------------------
#
def path_interval(curves, t0, t1):

    ic = []
    for c in curves:
        if c[-1,0] < t0:
            continue
        i0 = i1 = None
        for i,(t,px,py,vx,vy,w) in enumerate(c):
            if t >= t0:
                if i0 is None:
                    i0 = i1 = i
                elif t <= t1:
                    i1 = i
        if not i0 is None:
            ic.append(c[i0:i1+1,:].copy())
    return ic

# ------------------------------------------------------------------------------
#
def curve_closest_point(xy, curves):
    x,y, = xy
    d2min = ti = None
    for c in curves:
        for t,px,py,vx,vy,w in c:
            dx,dy = px-x, py-y
            d2 = dx*dx + dy*dy
            if d2min is None or d2 < d2min:
                d2min = d2
                ti = t
    return ti

# ------------------------------------------------------------------------------
#
def test():
    import sys
    if len(sys.argv) != 5:
        print ("bounce.py <angle> <vx> <vy> <h>")
        return
    params = [float(x) for x in sys.argv[1:]]
    from math import pi
    pa = pi * params[0] / 180.0
    vp = params[1:3]
    hs = params[3]
    serve(vp, pa, hs)

#test()

