// -----------------------------------------------------------------------------
//
#include <math.h>			// Use sqrt(), M_PI, fabs()
#include <Python.h>			// use PyObject

#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>		// use PyArray_*(), NPY_*

// -----------------------------------------------------------------------------
//
static double enters_circle(double x0, double y0, double x1, double y1,
			    double cx, double cy, double r)
{
  double r0 = sqrt((x0-cx)*(x0-cx) + (y0-cy)*(y0-cy));
  double r1 = sqrt((x1-cx)*(x1-cx) + (y1-cy)*(y1-cy));
  if (r0 >= r && r1 < r)
    {
      double f = (r0-r)/(r0-r1);
      return f;
    }
  return -1;
}

// -----------------------------------------------------------------------------
// Compute parabolic flight of ball until table strike (y=th "table height")
// within +/-0.5*tl (half table length), otherwise compute floor strike.
//
static 
const char* flight(double t0, double p0x, double p0y,
		   double v0x, double v0y, double w0,
		   double r, double mb, double th, double tl, double nh,
		   double cdrag, double clift, double cspin0, double cspinv,
		   double airdens, double airvisc, double g, double tstep,
		   int maxstep, int pmax, double *path, int *plen)
{
  int pi = 0;
  if (pmax > 0)
    {
      path[0] = t0; path[1] = p0x; path[2] = p0y;
      path[3] = v0x; path[4] = v0y; path[5] = w0;
      pi += 1;
    }
  if (p0y < r)
    {
      *plen = (pmax > 0 ? 1 : 0);
      return "floor";        // Under floor
    }

  double A = M_PI*r*r;
  double fdrag = cdrag*0.5*airdens*A/mb;
  double flift = clift*0.5*airdens*A/mb;

  double wdrag0 = -cspin0*sqrt(airvisc*airdens)*r*r*1.5/mb;
  double wdragv = -cspinv*1.5*A*airdens/mb;

  const char *hit = NULL;
  double t = t0, px = p0x, py = p0y, vx = v0x, vy = v0y, w = w0;
  for (int s = 0 ; hit == NULL && s < maxstep ; ++s)
    {
      double vmag = sqrt(vx*vx + vy*vy);
      double ax = -fdrag*vmag*vx - flift*w*r*vy;
      double ay = -fdrag*vmag*vy + flift*w*r*vx - g;
      double t1 = t + tstep;
      double p1x = px+vx*tstep+0.5*ax*tstep*tstep;
      double p1y = py+vy*tstep+0.5*ay*tstep*tstep;
      double v1x = vx+ax*tstep;
      double v1y = vy+ay*tstep;
      double wdt = wdrag0*w*sqrt(fabs(w)) + wdragv*vmag*w;
      double w1 = w + wdt*tstep;
      double dt = 0;
      if (py >= th+r && p1y < th+r)
	{
	  dt = tstep * (th+r - py) / (p1y - py);
	  if (fabs(px+vx*dt) <= 0.5*tl)
	    hit = "table";
	}
      if (py >= r && p1y < r && hit == NULL)
	{
	  dt = tstep * (r - py) / (p1y - py);
	  hit = "floor";
	}
      if ((px <= -r && p1x > -r) || (px > r && p1x <= r))
	{
	  double hx = (px <= -r ? -r : r);
	  double cf = (px - hx) / (px - p1x);
	  double y = (1-cf)*py + cf*p1y;
	  if (y >= th && y <= th+nh && (hit == NULL || tstep*cf < dt))
	    {
	      dt = tstep * cf;
	      hit = "net";
	    }
	}
      if (hit == NULL)
	{
	  double cf = enters_circle(px, py, p1x, p1y, 0, th+nh, r);
	  if (cf >= 0)
	    {
	      dt = tstep * cf;
	      hit = "net edge";
	    }
	}
      if (hit)
	{
	  t1 = t + dt;
	  p1x = px+vx*dt+0.5*ax*dt*dt;
	  p1y = py+vy*dt+0.5*ay*dt*dt;
	  v1x = vx+ax*dt;
	  v1y = vy+ay*dt;
	  w1 = w + wdt*dt;
	}


      if (hit && pi == pmax && pmax > 0)
	pi -= 1;
      if (pi < pmax)
	{
	  double *s = path + 6*pi;
	  s[0] = t1; s[1] = p1x; s[2] = p1y; s[3] = v1x; s[4] = v1y, s[5] = w;
	  pi += 1;
	}
      t = t1; px = p1x; py = p1y; vx = v1x; vy = v1y; w = w1;
    }
  *plen = pi;

  return (hit ? hit : "");
}

// ----------------------------------------------------------------------------
//
extern "C" PyObject *flight_py(PyObject *, PyObject *args, PyObject *keywds)
{
  double t0, p0x, p0y, v0x, v0y, w0, r, mb, th, tl, nh;
  double cdrag, clift, cspin0, cspinv, airdens, airvisc, g, tstep;
  int maxstep;
  PyObject *path;
  const char *kwlist[] = {"t0", "p0x", "p0y", "v0x", "v0y", "w0",
			  "r", "mb", "th", "tl", "nh",
			  "cdrag", "clift", "cspin0", "cspinv",
			  "airdens", "airvisc", "g", "tstep", "maxstep", "path",
			  NULL};
 if (!PyArg_ParseTupleAndKeywords(args, keywds,
				  const_cast<char *>("dddddddddddddddddddiO"),
				  (char **)kwlist,
				  &t0, &p0x, &p0y, &v0x, &v0y, &w0,
				  &r, &mb, &th, &tl, &nh,
				  &cdrag, &clift, &cspin0, &cspinv,
				  &airdens, &airvisc, &g, &tstep, &maxstep,
				  &path))
   return NULL;

 if (!PyArray_Check(path))
   {
     PyErr_SetString(PyExc_TypeError, "path argument must be a NumPy array");
     return NULL;
   }
 PyArrayObject *pa = (PyArrayObject *)path;
 if (PyArray_NDIM(pa) != 2 || PyArray_DIMS(pa)[1] != 6 ||
     PyArray_DESCR(pa)->type_num != NPY_DOUBLE ||
     PyArray_STRIDE(pa, 1) != sizeof(double) ||
     PyArray_STRIDE(pa, 0) != 6*sizeof(double))
   {
     PyErr_SetString(PyExc_TypeError, "path argument must be contiguous 2-dimensional N by 6 array");
     return NULL;
   }
 int pmax = PyArray_DIMS(pa)[0];
 double *p = static_cast<double *>(PyArray_DATA(pa));

 int n;
 const char *hit = flight(t0, p0x, p0y, v0x, v0y, w0, r, mb, th, tl, nh,
			  cdrag, clift, cspin0, cspinv,
			  airdens, airvisc, g, tstep, maxstep, pmax, p, &n);

 PyObject *t = PyTuple_New(2);
 PyTuple_SetItem(t, 0, PyUnicode_FromString(hit));
 PyTuple_SetItem(t, 1, PyLong_FromLong(n));

 return t;
}

// ----------------------------------------------------------------------------
//
static struct PyMethodDef fly_methods[] =
{
  {const_cast<char*>("flight"), (PyCFunction)flight_py,
   METH_VARARGS|METH_KEYWORDS, NULL},
  {NULL, NULL, 0, NULL}
};

static struct PyModuleDef fly_def =
{
	PyModuleDef_HEAD_INIT,
	"fly",
	"Optimize flight equation solver",
	-1,
	fly_methods,
	NULL,
	NULL,
	NULL,
	NULL
};

// ----------------------------------------------------------------------------
// Initialization routine called by python when module is dynamically loaded.
//
PyMODINIT_FUNC PyInit_fly()
{
  import_array();	// required before using NumPy.

 return PyModule_Create(&fly_def);
}
