from math import pi, sin, cos, atan2, exp, sqrt, ceil

from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import Qt

from . import bounce

# Serve parameters: x, y, tilt, speed, swing
serves = (('short back spin', (.5, .3, 70.0, 12.0, -21.0)),
          ('short no spin', (.5, .28, 30.0, 7.0, -43.0)),
          ('short top spin', (.5, .28, 10.0, 3.0, 0.0)),
          ('long back spin', (.5, .26, 50.0, 14.0, -35)),
          ('long no spin', (.5, .23, 20.0, 8.0, -43.0)),
          ('long top spin', (.5, .17, -20.0, 12.0, 32.0)))

from PyQt5.QtWidgets import QMainWindow
class Ping_Pong(QMainWindow):

    def __init__(self, parent=None):

        QMainWindow.__init__(self, parent)
        self.setWindowTitle("SpinDoctor")
        from PyQt5.QtWidgets import QWidget, QGridLayout, QShortcut
        self.setCentralWidget(QWidget())

        # Simulation parameters
        from .bounce import Physical_Parameters, Ball_State
        self.physical_parameters = Physical_Parameters()
        self.time_step = 0.001

        self.in_play = True
        self.side = 'right'
        self.serve = True
        self.strike = True
        self.toss_height = 0.30
        self.incoming_ball = Ball_State(0, (1.7,1.0), (0,0), 0)
        self.paddle_angle = 173 * pi/180        # radians
        self.paddle_velocity = (-3.5,-2)        # meters per second
        self.outgoing_ball = Ball_State()
        self.received_ball = Ball_State()
        self.receive_deep = False
        self.last_flight_params = None
        self.outgoing_path = []

        # Drawing parameters
        self.scale = 200.0                # pixels per meter
        self.speed_scale = 0.1            # fraction of scale.
        self.rotation_speed_scale = 5.0   # meters per second per turn
        self.xoffset = 0         # pixels, center of table from window midline
        self.time0 = None        # clock time base to animate ball motion
        self.pause_time = None
        self.rally = False
        self.show_paddle_velocity = True
        self.show_spin = True
        self.swing_duration = 0.15      # seconds, harmonic motion
        self.slow_motion = 1.0

        from PyQt5.QtGui import QColor
        self.graphics_background = QColor(250, 250, 200)        # Light yellow

        self.table_thickness = 0.0254 * 1.5
        self.table_color = QColor(150,150,255)        # blue
        self.net_color = Qt.black
        self.leg_thickness = 0.0254 * 2
        self.leg_position = 0.8         # fraction of half table length
        self.table_leg_color = Qt.black
        self.net_thickness = 0.0254 * 0.5       # for display

        self.paddle_thickness = 0.02
        self.paddle_length = 0.15
        self.paddle_color = Qt.red

        self.ball_color = QColor(255,179,0) # Orange
        self.spin_color = Qt.white
        self.spin_arrow_width = 0.01

        self.velocity_arrow_width = 0.02
        self.velocity_arrow_color = Qt.white

        # Parameter display
        self.show_params = True
        self.show_ranges = False
        self.param_ranges = None
        self.show_const = False

        # Velocity scan.
        self.show_vscan = False
        vstep = 0.5
        self.vscan_range = {'left': ((-5,20,vstep),(-10,20,vstep)),
                            'right': ((-20,5,vstep),(-10,20,vstep))}
        self.vscan_colors = (QColor(170,255,170),       # in
                             QColor(250,250,190),       # short
                             QColor(255,170,170))       # long
        self.vscan = None
        self.vscan_image = None

        self.markers = []

        # Mouse drag state.
        self.drag_type = None
        self.drag_position = None
        self.dragged = False

        # Mouse wheel state.
        self.wheel_mode = None
        self.wheel_xy = None
        self.paddle_angle_wheel_step = 5        # degrees
        self.spin_wheel_step = 1                # m/s

        # Layout main window.

        g = QGridLayout()

        p, self.fields = self.parameter_panel()
        g.addWidget(p, 0, 0, 2, 1)
        self.parameters = p
        self.display_parameters(self.show_params)

        gr = self.graphics = Ping_Pong_Graphics(self)
        g.addWidget(gr, 0, 1)
        g.setColumnStretch(1, 10)

        c, self.cfields = self.constants_panel()
        c.hide()
        g.addWidget(c, 0, 2, 2, 1)
        self.constants = c

        g.setSpacing(0)
        g.setContentsMargins(0,0,0,0)
        self.centralWidget().setLayout(g)

        self.create_menus()

        tb = self.create_toolbar()
        g.addWidget(tb, 1, 1, Qt.AlignHCenter)

        close_key = QtGui.QKeySequence(Qt.CTRL + Qt.Key_Q)
        QShortcut(close_key, self, self.close)

    def parameter_panel(self):

        from PyQt5.QtWidgets import QGroupBox, QGridLayout, QLabel, QLineEdit
        p = QGroupBox(self.centralWidget())
        p.setBackgroundRole(QtGui.QPalette.Base)
        p.setAutoFillBackground(True)
        lo = QGridLayout()
        params = ('Incoming ball',
                  ('toss', 'm'), ('speed', 'm/s'), ('slope', 'deg'),
                  ('spin', 'm/s'), ('height', 'm'), ('depth', 'm'),
                  'Hit',
                  ('tilt', 'deg'), ('speed', 'm/s'), (' flat', 'm/s'),
                  (' slice', 'm/s'), ('swing', 'deg'), ('grip', ''), ('time', 's'),
                  'Outgoing ball',
                  ('speed', 'm/s'), ('slope', 'deg'), ('spin', 'm/s'),
                  'Return',
                  ('time', 's'), ('speed', 'm/s'), ('slope', 'deg'),
                  ('spin', 'm/s'), ('height', 'm'), ('depth', 'm'),
                  ('drag', 'g'), ('lift', 'g')
                  )
        f = QtGui.QFont("Courier", 14, QtGui.QFont.Normal)
        m = QtGui.QFontMetrics(f)
        fs = m.size(Qt.TextSingleLine, '-888.8')
        s = QtCore.QSize(fs.width(), fs.height()+2)
        fields = {}
        for r, param in enumerate(params):
            if isinstance(param, tuple):
                name,units = param
                n = QLabel(' ' + name)
                n.setFont(f)
                lo.addWidget(n, r, 0)
                e = QLineEdit()
                e.setFixedSize(s)
                e.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
                e.setFont(f)
                e.setFrame(False)
#                e.setBackgroundRole(QtGui.QPalette.Window)     # No effect.
                def c(s = self, h = heading, n = name.strip()):
                    s.parameter_changed(h, n)
                e.returnPressed.connect(c)
                lo.addWidget(e, r, 1)
                u = QLabel(units)
                u.setFont(f)
                lo.addWidget(u, r, 2)
                rn = QLabel()      # Range
                rn.setFont(f)
                rn.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
                rn.setFixedWidth(int(2.5*s.width()))
                lo.addWidget(rn, r, 3)
                fields[(heading,name.strip())] = e, rn

            else:
                heading = param
                h = QLabel(heading)
                h.setFont(f)
                h.setFixedHeight(2*s.height())
                lo.addWidget(h, r, 0, 1, 3)

        lo.setVerticalSpacing(0)
        lo.setHorizontalSpacing(10)
        p.setLayout(lo)
        lo.setRowStretch(r+1, 10)
        lo.setContentsMargins(3,3,3,3)

        return p, fields

    def parameter_changed(self, heading, name):

        text = self.fields[(heading,name)][0].text()
        value = float(text)
        sd = -1 if self.side == 'right' else 1
        pp = self.physical_parameters
        if heading == 'Incoming ball':
            ib = self.incoming_ball
            if name == 'toss':
                self.toss_height = value
            elif name == 'speed':
                ib.velocity = normalize(ib.velocity, value, (-sd,0))
            elif name == 'slope':
                a = value*pi/180.0
                n = norm(ib.velocity)
                ib.velocity = (-sd*n*cos(a), n*sin(a))
            elif name == 'spin':
                ib.spin = sd*value / pp.ball_radius
            elif name == 'height':
                h = pp.table_height + value
                ib.position = (ib.position[0], h)
            elif name == 'depth':
                tlh = 0.5*pp.table_length
                x = -sd * (tlh + value)
                ib.position = (x, ib.position[1])
            self.strike = True
        elif heading == 'Hit':
            if name == 'tilt':
                a = value*pi/180
                self.paddle_angle = atan2(sin(a),sd*cos(a))
            elif name == 'speed':
                self.paddle_velocity = normalize(self.paddle_velocity, value,
                                                 (sd,0))
            elif name == 'flat':
                a = self.paddle_angle
                c, s = cos(a), sin(a)
                pv = self.paddle_velocity
                ps = -s*pv[0] + c*pv[1]
                self.paddle_velocity = (value*c - ps*s, value*s + ps*c)
            elif name == 'slice':
                a = self.paddle_angle
                c, s = cos(a), sin(a)
                pv = self.paddle_velocity
                pf = c*pv[0] + s*pv[1]
                self.paddle_velocity = (pf*c - sd*value*s, pf*s + sd*value*c)
            elif name == 'swing':
                a = value*pi/180.0
                pv = self.paddle_velocity
                n = norm(pv)
                self.paddle_velocity = (sd*n*cos(a), n*sin(a))
            self.strike = True
        elif heading == 'Outgoing ball':
            ob = self.outgoing_ball
            if name == 'speed':
                ob.velocity = normalize(ob.velocity, value, (sd,0))
            elif name == 'slope':
                a = value*pi/180.0
                n = norm(ob.velocity)
                ob.velocity = (sd*n*cos(a), n*sin(a))
            elif name == 'spin':
                ob.spin = -sd*value / pp.ball_radius
            ib = self.incoming_ball
            from .bounce import calculate_strike
            try:
                self.paddle_angle, self.paddle_velocity = \
                    calculate_strike(ib.velocity, ib.spin, ob.velocity, ob.spin,
                                     pp.paddle_friction, pp.paddle_damping,
                                     pp.ball_radius)
                self.strike = True
            except ValueError:
                self.strike = False        # Unachievable spin
                
        self.redraw()


    def constants_panel(self):

        from PyQt5.QtWidgets import QGroupBox, QGridLayout, QLabel, QLineEdit
        p = QGroupBox(self.centralWidget())
        p.setBackgroundRole(QtGui.QPalette.Base)
        p.setAutoFillBackground(True)
        lo = QGridLayout()
        params = ('Paddle',
                  ('friction', '%.2f', 'paddle_friction', ''),
                  ('damping', '%.2f', 'paddle_damping', ''),
                  'Ball',
                  ('radius', '%.3f', 'ball_radius', 'm'),
                  ('mass', '%.4f', 'ball_mass', 'kg'),
                  'Table',
                  ('height', '%.3f', 'table_height', 'm'),
                  ('length', '%.3f', 'table_length', 'm'),
                  ('friction', '%.2f', 'table_friction', ''),
                  ('damping', '%.2f', 'table_damping', ''),
                  ('net height', '%.3f', 'net_height', 'm'),
                  'Flight',
                  ('drag', '%.2f', 'drag_coefficient', ''),
                  ('lift', '%.2f', 'lift_coefficient', ''),
                  ('spin decay', '%.3f', 'spin_drag_coefficient', ''),
                  ('spin down', '%.2f', 'spin_down_coefficient', ''),
                  ('air density', '%.2f', 'air_density', 'kg/m**3'),
                  ('air viscosity', '%.3g', 'air_viscosity', 'kg/m*s'),
                  ('gravity', '%.2f', 'gravity', 'm/s**2'),
                  ('time step', '%.4f', 'time_step', 's'),
                  )
        f = QtGui.QFont("Courier", 14, QtGui.QFont.Normal)
        m = QtGui.QFontMetrics(f)
        fs = m.size(Qt.TextSingleLine, '-888.88')
        s = QtCore.QSize(fs.width(), fs.height()+2)
        fields = {}
        for r, param in enumerate(params):
            if isinstance(param, tuple):
                name,format,pname,units = param
                n = QLabel(' ' + name)
                n.setFont(f)
                lo.addWidget(n, r, 0)
                e = QLineEdit()
                e.setFixedSize(s)
                e.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
                e.setFont(f)
                e.setFrame(False)
                def c(s = self, h = heading, n = name.strip()):
                    s.constant_changed(h, n)
                e.returnPressed.connect(c)
                lo.addWidget(e, r, 1)
                u = QLabel(units)
                u.setFont(f)
                lo.addWidget(u, r, 2)
                fields[(heading,name.strip())] = (e, format, pname)
            else:
                heading = param
                h = QLabel(heading)
                h.setFont(f)
                h.setFixedHeight(2*s.height())
                lo.addWidget(h, r, 0, 1, 3)

        lo.setVerticalSpacing(0)
        lo.setHorizontalSpacing(10)
        p.setLayout(lo)
        lo.setRowStretch(r+1, 10)
        lo.setContentsMargins(3,3,3,3)

        return p, fields

    def show_constants(self):

        pp = self.physical_parameters
        f = self.cfields
        for (e,format,pname) in  f.values():
            if hasattr(pp, pname):
                e.setText(format % getattr(pp, pname))
            elif pname == 'time_step':
                e.setText(format % self.time_step)

    def constant_changed(self, heading, name):

        e,format,pname = self.cfields[(heading,name)]
        val = float(e.text())
        pp = self.physical_parameters
        if hasattr(pp, pname):
            setattr(pp, pname, val)
        elif pname == 'time_step':
            self.time_step = val
        self.redraw()

    def create_menus(self):

        # TODO: QMainWindows docs say  application menu name is taken from
        #       Info.plist.  Apparently cannot be set in PySide.

        vm = self.menuBar().addMenu("&View")

        def dp(checked, self=self):
            self.display_parameters(self.dpmenu.isChecked())
        from PyQt5.QtWidgets import QAction
        self.dpmenu = QAction("Show parameters", self,
                              checkable = True,
                              checked = self.show_params,
                              shortcut=Qt.CTRL + Qt.Key_P,
                              triggered=dp)
        # TODO: QAction toggled signal not recognized by pyside.
        vm.addAction(self.dpmenu)

        def dpr(checked, self=self):
            self.display_parameter_ranges(self.dprmenu.isChecked())
        self.dprmenu = QAction("Show parameter ranges", self,
                               checkable = True, checked = False,
                               shortcut=Qt.CTRL + Qt.Key_I,
                               triggered=dpr)
        vm.addAction(self.dprmenu)

        def dpv(checked, self=self):
            self.show_vscan = self.dpvmenu.isChecked()
            self.redraw()
        self.dpvmenu = QAction("Show in-play paddle velocities", self,
                               icon = self.icon('vscan.png'),
                               iconText = 'in-play',
                               checkable = True, checked = True,
                               shortcut=Qt.CTRL + Qt.Key_B,
                               triggered=dpv)
        vm.addAction(self.dpvmenu)

        def dc(checked, self=self):
            self.display_constants(self.dcmenu.isChecked())
        self.dcmenu = QAction("Show constants", self,
                              checkable = True, checked = False,
                              shortcut=Qt.CTRL + Qt.Key_K,
                              triggered=dc)
        vm.addAction(self.dcmenu)

        def dpa(checked, self=self):
            self.show_paddle_velocity = self.dpamenu.isChecked()
            self.redraw()
        self.dpamenu = QAction("Show paddle velocity arrow", self,
                               icon = self.icon('parrow.png'),
                               iconText = 'arrow',
                               checkable = True,
                               checked = self.show_paddle_velocity,
                               shortcut=Qt.CTRL + Qt.Key_W,
                               triggered=dpa)
        self.dpamenu.setChecked(self.show_paddle_velocity)
        vm.addAction(self.dpamenu)

        def ds(checked, self=self):
            self.show_spin = self.dsmenu.isChecked()
            self.redraw()
        self.dsmenu = QAction("Show spin arrow", self,
                              checkable = True,
                              checked = self.show_spin,
                              icon = self.icon('spin.png'),
                              iconText = 'spin',
                              shortcut=Qt.CTRL + Qt.Key_N,
                              triggered=ds)
        self.dsmenu.setChecked(self.show_spin)
        vm.addAction(self.dsmenu)

        sm = self.menuBar().addMenu('&Serve')

        for sname, sparam in serves:
            ts = lambda checked, s=self, sn=sname: s.named_serve(sn)
            se = QAction(sname, self, triggered=ts)
            sm.addAction(se)

        self.smenu = QAction('Serve', self,
                             icon = self.icon('serve.png'),
                             iconText = 'serve',
                             triggered = lambda checked, s=self: s.named_serve('short back spin'),
                             shortcut=Qt.CTRL + Qt.Key_S)

        am = self.menuBar().addMenu('&Play')

        m = QAction("Show motion", self,
                    shortcut=Qt.CTRL + Qt.Key_M,
                    triggered=lambda checked, self=self: self.show_motion())
        am.addAction(m)

        def sm(checked, self=self):
            slomo = 0.2 if self.smmenu.isChecked() else 1.0
            if not self.time0 is None:
                from time import time
                tnow = time() if self.pause_time is None else self.pause_time
                self.time0 = tnow - (tnow - self.time0)*self.slow_motion/slomo
            self.slow_motion = slomo
        self.smmenu = QAction("Slow motion", self,
                              iconText = 'Slow',
                              checkable = True, checked = False,
                              shortcut=Qt.CTRL + Qt.Key_L,
                              triggered=sm)
        am.addAction(self.smmenu)

        m = QAction("Hit return", self,
                    shortcut=Qt.CTRL + Qt.Key_R,
                    triggered=lambda checked, self=self: self.hit_return())
        am.addAction(m)

        self.ramenu = QAction("Rally", self,
                              shortcut=Qt.CTRL + Qt.Key_6,
                              triggered=lambda checked, self=self: self.toggle_rally())
        am.addAction(self.ramenu)

        def rd(checked, self=self):
            self.receive_deep = self.rdmenu.isChecked()
            self.set_receive_position()
            self.redraw()
        self.rdmenu = QAction("Receive deep", self,
                              checkable = True,
                              checked = self.receive_deep,
                              shortcut=Qt.CTRL + Qt.Key_D,
                              triggered=rd)
        am.addAction(self.rdmenu)

        hm = self.menuBar().addMenu('Help')

        m = QAction("Show manual", self,
                    triggered=lambda checked, self=self: self.show_manual())
        hm.addAction(m)

    def create_toolbar(self):

        from PyQt5.QtWidgets import QToolBar, QAction
        tb = QToolBar()
        tb.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)

        sr = QAction("Start rally", self,
                     icon = self.icon('rally.png'),
                     iconText = 'rally',
                     triggered = lambda checked, self=self: self.start_rally(),
                     shortcut=Qt.CTRL + Qt.Key_R)
        tb.addAction(sr)

        pr = QAction("Pause rally", self,
                     icon = self.icon('pause.png'),
                     iconText = 'pause',
                     triggered = lambda checked, self=self: self.pause_rally(),
                     shortcut=Qt.CTRL + Qt.Key_P)
        tb.addAction(pr)

        st = QAction("Stop rally", self,
                     icon = self.icon('stop.png'),
                     iconText = 'stop',
                     triggered = lambda checked, self=self: self.stop_rally(),
                     shortcut=Qt.CTRL + Qt.Key_S)
        tb.addAction(st)

        hr = QAction("Hit return", self,
                     icon = self.icon('return.png'),
                     iconText = 'return',
                     triggered = lambda checked, self=self: self.hit_return(),
                     shortcut=Qt.CTRL + Qt.Key_R)

        tb.addAction(self.smmenu)       # speed

        tb.addSeparator()

        tb.addWidget(self.serve_tool_button())
        tb.addAction(hr)                # hit return


        mo = QAction("Show ball motion", self,
                     icon = self.icon('motion.png'),
                     iconText = 'motion',
                     triggered = lambda checked, self=self: self.show_motion(),
                     shortcut=Qt.CTRL + Qt.Key_M)
        tb.addAction(mo)

        tb.addSeparator()

        tb.addAction(self.dpamenu)      # paddle arrow
        tb.addAction(self.dsmenu)       # spin arrow
        tb.addAction(self.dpvmenu)      # in-play velocities

        return tb

    def show_manual(self):
        from os.path import dirname, join
        url = 'file://%s' % join(dirname(__file__), 'docs', 'spin.html')
        import webbrowser
        webbrowser.open(url)

    def serve_tool_button(self):

        from PyQt5.QtWidgets import QToolButton, QMenu, QAction, QActionGroup
        stb = QToolButton()
        stb.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        sm = QMenu()
        g = QActionGroup(self, exclusive = True)
        for t in ('short back spin', 'short no spin', 'short top spin',
                  'long back spin', 'long no spin', 'long top spin',
                  'middle back spin', 'middle no spin', 'middle top spin'):
            a = QAction(t, self, checkable = True,
                        checked = (t == 'short back spin'),
                        triggered = lambda checked, s=self, t=t: s.named_serve(t))
            a.setActionGroup(g)
            sm.addAction(a)
        stb.setPopupMode(QToolButton.MenuButtonPopup)
        stb.setMenu(sm)
        st = lambda checked, s=self, g=g: s.named_serve(g.checkedAction().text())
        sa = QAction('Serve', self,
                     icon = self.icon('serve.png'),
                     iconText = 'serve',
                     triggered = st)
        stb.setDefaultAction(sa)
        return stb

    def icon(self, filename):
        from os.path import join, dirname
        path = join(dirname(__file__), 'icons', filename)
        return QtGui.QIcon(path)

    def display_parameters(self, show = True):

        self.show_params = show
        p = self.parameters
        if show:
            self.show_parameters()
            p.show()
        else:
            p.hide()

    def display_parameter_ranges(self, show = True):

        self.show_ranges = show
        if show:
            self.display_parameters(True)
        else:
            self.show_parameters()

    def display_constants(self, show = True):

        self.show_const = show
        c = self.constants
        if show:
            self.show_constants()
            c.show()
        else:
            c.hide()

    def mouse_press(self, event):

        b = event.button()
        if b != Qt.LeftButton:
            return

        self.wheel_mode = None
        p = event.pos()
        if self.paddle_velocity_click(p):
            t = 'velocity'
        elif self.ball_click(p, self.top_of_toss()):
            t = 'ball'
        elif self.paddle_center_click(p):
            t = 'move'
        elif self.paddle_click(p):
            t = 'angle'
        elif self.ball_click(p, self.received_ball.position):
            t = 'ball marker'
        elif self.under_table_click(p):
            t = 'zoom'
        elif self.bounce_click(p):
            t = 'depth'
        elif self.over_net_click(p):
            t = 'height'
        else:
            t = 'velocity'
        self.drag_type = t
        self.drag_position = p
        self.dragged = False
        self.marker_click(p)
                
    def mouse_move(self, event):

        self.reset_wheel_mode(event)
        if self.drag_position is None or self.drag_type is None:
            return

        t = self.drag_type
        pos = event.pos()
        speed = 0.1 if event.modifiers() & Qt.ShiftModifier else 1.0
        lp = self.drag_position
        s = self.scale
        dx,dy = speed*(pos.x()-lp.x())/s, speed*(lp.y()-pos.y())/s
        if t == 'angle':
            x,y = self.paddle_center()
            dx, dy = pos.x() - x, y - pos.y()
            dx0, dy0 = lp.x() - x, y - lp.y()
            da = atan2(dy,dx) - atan2(dy0,dx0)
            da = (da + pi) % (2*pi) - pi
            self.paddle_angle += speed * da
            self.redraw()
        elif t == 'move':
            ib = self.incoming_ball
            # Stay on incoming curve if not serving.
            if self.serve:
                sx,sy = ib.position
                ib.position = (sx + dx, sy + dy)
            else:
                path = [self.incoming_ball.incoming_curve]
                p = self.pixel_to_xy((pos.x(), pos.y()))
                t = bounce.curve_closest_point(p, path)
                ib = self.incoming_ball
                bounce.set_ball_state_at_time(ib, t, [ib.incoming_curve])
            self.redraw()
        elif t == 'ball':
            self.toss_height += dy
            ib = self.incoming_ball
            sx,sy = ib.position
            ib.position = (sx + dx, sy)
            self.redraw()
        elif t == 'velocity':
            s = self.speed_scale
            vx,vy = self.paddle_velocity
            self.paddle_velocity = (vx + dx/s, vy + dy/s)
            self.redraw()
        elif t == 'ball marker':
            p = self.pixel_to_xy((pos.x(), pos.y()))
            t = bounce.curve_closest_point(p, self.outgoing_path)
            self.set_receive_time(t, self.outgoing_path)
            self.redraw()
        elif t == 'zoom':
            self.scale *= exp(-speed*dy*self.scale/100)
            self.xoffset += speed*dx*self.scale
            self.redraw()
        elif t == 'depth':
            sd = -1 if self.side == 'right' else 1
            self.adjust_flight(depth = dx*sd)
            self.redraw()
        elif t == 'height':
            self.adjust_flight(height = dy)
            self.redraw()
        self.drag_position = pos
        self.dragged = True

    def mouse_release(self, event):

        if event.button() == Qt.LeftButton:
            if self.drag_type == 'ball marker' and not self.dragged:
                self.hit_return()
            self.drag_type = None
            self.drag_position = None

    def mouse_wheel(self, event):

        self.reset_wheel_mode(event)
        mode = self.wheel_mode
        if mode is None:
            p = event.pos()
            if self.ball_click(p, self.received_ball.position):
                mode = 'spin'
            else:
                mode = 'paddle angle'
            self.wheel_mode = mode
            self.wheel_xy = (p.x(),p.y())

        d = event.angleDelta().y()       # in 1/8 degree, typically 120 multiples.
        sd = -1 if self.side == 'right' else 1
        da = d/120.0
        speed = 0.1 if event.modifiers() & Qt.ShiftModifier else 1.0
        if mode == 'spin':
            r = self.physical_parameters.ball_radius
            self.adjust_flight(spin = self.spin_wheel_step/r*speed*da*-sd)
        elif mode == 'paddle angle':
            astep = speed * self.paddle_angle_wheel_step * pi/180 * -sd * da
            self.paddle_angle += astep
        self.redraw()

    def reset_wheel_mode(self, event):

        p = event.pos()
        ex, ey = p.x(),p.y()
        if self.wheel_xy:
            wx, wy = self.wheel_xy
            if abs(ex-wx) > 10 or abs(ey-wy) > 10:
                self.wheel_mode = None

    #
    # Adjust paddle angle and velocity to change depth of bounce, height over
    # net and post-bounce spin.
    #
    def adjust_flight(self, depth = 0, height = 0, spin = 0):

        h, d, s = bounce.receive_parameters(self.outgoing_path,
                                            self.serve, self.side)
        if h is None or d is None or s is None:
            return
        ib = self.incoming_ball
        pa, vp = bounce.hit_target(h + height, d + depth, s + spin,
                                   ib.time, ib.position, ib.velocity, ib.spin,
                                   self.paddle_angle, self.paddle_velocity,
                                   self.side, self.serve,
                                   self.physical_parameters, self.time_step)
        self.paddle_angle = pa
        self.paddle_velocity = vp

    def named_serve(self, serve_name):

        for sname, sparam in serves:
            if sname == serve_name:
                break
        self.serve = True
        self.strike = True
        ib = self.incoming_ball
        x, y, pa, ps, sa = sparam
        pa *= pi/180
        sa *= pi/180
        sd = -1 if self.side == 'right' else 1
        pp = self.physical_parameters
        tlh = 0.5*pp.table_length
        th = pp.table_height
        ib.position = (-sd*(x + tlh), y + th)
        ib.spin = 0
        self.paddle_angle = atan2(sin(pa),sd*cos(pa))
        self.paddle_velocity = (sd*ps*cos(sa), ps*sin(sa))
        self.redraw()

    def hit_return(self):

        ib = self.incoming_ball
        ib.set_state(self.received_ball)

        self.serve = False
        self.side = side = 'left' if self.side == 'right' else 'right'
        self.strike = True
        self.vscan = self.vscan_image = None

        sd = -1 if side == 'right' else 1
        pp = self.physical_parameters
        shots = self.return_parameters(ib)
        pa = 0
        vp = (0, 0)
        in_play = False
        for v,a,s in shots:
            wout = -sd*s/pp.ball_radius
            ar = a*pi/180.
            vx,vy = sd*v*cos(ar), v*sin(ar)
            vout = (vx,vy)
            from .bounce import calculate_strike
            try:
                pa, vp = calculate_strike(ib.velocity, ib.spin, vout, wout,
                                          pp.paddle_friction, pp.paddle_damping,
                                          pp.ball_radius)
            except ValueError:
                continue        # Unachievable spin

            in_play = self.hit_flight(side, False, 0, ib.time,
                                      ib.position, ib.velocity, ib.spin,
                                      pa, vp, path = False)
            if in_play:
                break

        self.paddle_angle = pa
        self.paddle_velocity = vp
        self.redraw()
        return in_play

    def return_parameters(self, incoming_ball):
        # Specify returns using outgoing (ball speed, ball angle, spin).

        #
        # TODO: Should choose most sensible shot type based on incoming
        #       ball.  Also should be able to choose receive depth based
        #       on table strike position, time of table strike since paddle
        #       strike, horizontal speed at table strike, bounce height,
        #       spin after table strike.
        #
        # High bounce no more than 3 feet back from edge -> smash.
        # Double bounce on table -> push or flip.
        # Long back spin -> loop.
        # Long fast top spin -> block or chop
        # Long medium speed top spin -> drive or loop or chop
        # Long high bounce from smash -> lob, block or chop
        #
        ib = incoming_ball
        pp = self.physical_parameters
        tl = pp.table_length
        x, y = ib.position
        depth = abs(x) - 0.5*tl
        height = y - pp.table_height
        vin = norm(ib.velocity)
        sd = -1 if self.side == 'right' else 1
        backspin = (ib.spin > 0) if self.side == 'right' else (ib.spin < 0)
        from random import random
        if height >= 2*pp.net_height and depth <= 0.25*tl:
            # Smash
            name = 'smash'
            shots = ((20,-5,5), (20,-10,5), (20,-15,5), (20,0,5), (20,5,5),
                     (20,-20,5), (20,-30,5), (20,-40,5), (20,-50,5), (20,-60,5),
                     (20,-5,0), (20,-10,0), (20,-15,0), (20,0,0), (20,5,0),
                     (20,-20,0), (20,-30,0), (20,-40,0), (20,-50,0), (20,-60,0),
                     (15,-5,5), (15,-10,5), (15,-15,5), (15,0,5), (15,5,5))
        elif depth >= 0.7*tl:
            # Lob or chop
            if random() > 0.5:
                # Lob
                name = 'lob'
                shots = ((10,30,5), (10,40,5), (10,20,5), (10,25,5), (10,35,5),
                         (12,30,5), (12,40,5), (12,20,5), (12,25,5), (12,35,5),
                         (15,30,5), (15,40,5), (15,20,5), (15,25,5), (15,35,5))
            else:
                # Chop
                name = 'chop'
                shots = ((10,0,-5),(10,10,-5),(10,20,-5),(10,30,-5),
                         (8,10,-5),(8,20,-5),(8,30,-5),
                         (6,20,-5),(6,30,-5),(6,40,-5),
                         (4,20,-5),(4,30,-5))
        elif depth >= 0:
            if backspin or random() > 0.7:
                # Loop
                name = 'loop'
                shots = ((20,0,10),(20,5,10),(20,10,10),(20,15,10),
                         (15,0,10),(15,5,10),(15,10,10),(15,15,10),
                         (10,0,10),(10,10,10),(10,20,10),(10,30,10),
                         (10,0,5),(10,10,5),(10,20,5),(10,30,5),
                         (8,10,5),(8,20,5),(8,30,5),
                         (6,20,5),(6,30,5),(6,40,5),
                         (4,20,5),(4,30,5))
            elif vin >= 9:
                # Block
                name = 'fast block'
                shots = ((vin, 10, -3), (vin, 15, -3), (vin, 5, -3),
                         (vin, 20, -3), (vin, 25, -3), (vin, 30, -3))
            else:
                # Drive
                name = 'drive'
                shots = ((10,0,5),(10,10,5),(10,20,5),(10,30,5),
                         (8,10,5),(8,20,5),(8,30,5),
                         (6,20,5),(6,30,5),(6,40,5),
                         (4,20,5),(4,30,5))
        elif backspin:
            # Push or flip.
            if random() > 0.2:
                # Push
                name = 'push'
                shots = ((5,0,-5), (5,5,-5), (5,10,-5), (5,15,-5),
                         (6,0,-3), (6,5,-3), (6,10,-2), (6,15,-2))
            else:
                # Flip
                name = 'backspin flip'
                shots = ((8, 0, 5), (8, 5, 5), (8, 10, 5), (8, 15, 5),
                         (8, 0, 0), (8, 5, 0), (8, 10, 0), (8, 15, 0))
        elif vin >= 8:
            # Block
            name = 'short block'
            shots = ((vin, 5, -3), (vin, 10, -3), (vin, 15, -3))
        else:
            # Flip
            name = 'topspin flip'
            shots = ((8, 0, 5), (8, 5, 5), (8, 10, 5), (8, 15, 5),
                     (8, 0, 0), (8, 5, 0), (8, 10, 0), (8, 15, 0))

        # Shuffle.
        shots = [s for r,s in sorted([(random(), s) for s in shots])]

        return shots

    def redraw(self):

        self.graphics.update()

    def start_rally(self):

        if self.rally:
            if not self.pause_time is None:
                from time import time
                self.time0 += time() - self.pause_time
                self.pause_time = None
        else:
            self.rally = True
            self.show_motion()

    def stop_rally(self):

        if self.rally:
            self.rally = False
            self.time0 = None
            self.killTimer(self.timer_id)
            self.redraw()

    def pause_rally(self):

        if self.rally:
            if self.pause_time is None:
                from time import time
                self.pause_time = time()
            else:
                self.start_rally()

    def toggle_rally(self):
        
        if self.rally:
            self.stop_rally()
        else:
            self.start_rally()

    def show_motion(self):

        ts = (self.incoming_ball.time - self.swing_duration) / self.slow_motion
        from time import time
        self.time0 = time() - ts
        tstep = 33      # milliseconds
        self.timer_id = self.startTimer(tstep)
        self.pause_time = None

    def timerEvent(self, event):

        self.redraw()

    def paint(self, event):

        # Recompute trajectory if needed.
        if self.flight_changed():
            if self.flight_changed(ignore_paddle_velocity = True):
                self.vscan = self.vscan_image = None
            self.calculate_flight()
            self.param_ranges = None

        pt = QtGui.QPainter(self.graphics)

        # Check if drawing ball in motion
        mo = None
        if not self.time0 is None:
            from time import time
            tnow = time() if self.pause_time is None else self.pause_time
            t = (tnow - self.time0) * self.slow_motion
            if self.rally:
                if t >= self.receive_bounce_time():
                    self.hit_return()
                    self.calculate_flight()
            mo = bounce.path_time_point(self.flight_path(), t)
            if mo is None:
                self.killTimer(self.timer_id)
                self.time0 = None
                self.rally = False

        # Draw in-play paddle velocities
        if self.show_vscan and self.strike:
            self.paint_vscan(pt)

        # Draw table
        self.paint_table(pt)

        # Draw paddle
        if self.strike:
            if mo is None:
                po = (0,0)
            else:
                v = self.paddle_velocity
                Ts = self.swing_duration
                t0 = t - self.strike_time()
                toff = (Ts/pi) * sin(pi*min(0.5,max(-1,t0/Ts)))
                po = (toff*v[0],toff*v[1])
            self.paint_paddle(pt, po)
            if self.show_paddle_velocity:
                self.paint_paddle_velocity(pt)

        # Draw flight path
        if mo is None:
            self.draw_curves(pt, self.flight_path(), scale=self.scale)

        # Draw balls
        pp = self.physical_parameters
        if mo is None:
            ib = self.incoming_ball
            if self.serve and self.strike:
                self.paint_ball(pt, self.top_of_toss(), ib.spin)
            self.paint_ball(pt, ib.position, ib.spin)
            rb = self.received_ball
            self.paint_ball(pt, rb.position, rb.spin)
        else:
            self.paint_ball(pt, (mo[1],mo[2]), mo[5])

        # Show parameter values
        if self.show_params:
            self.show_parameters()

        if len(self.markers) > 0:
            # paint markers
            pt.setPen(Qt.NoPen)
            pt.setBrush(Qt.green)
            r = 5
            cx, cy = self.paddle_center(pixels = False)
            s = self.speed_scale
            for a,vx,vy in self.markers:
                x,y = self.xy_to_pixel((cx + s*vx, cy + s*vy))
                pt.drawEllipse(x-r, y-r, 2*r, 2*r)

    def show_parameters(self):

        pp = self.physical_parameters
        sd = -1 if self.side == 'right' else 1
        vp = (vx,vy) = self.paddle_velocity
        ps = norm(vp)
        pa = self.paddle_angle
        pad = side_angle(pa,sd)
        cpa,spa = cos(pa), sin(pa)
        vf = cpa*vx + spa*vy
        vs = sd*(-spa*vx + cpa*vy)
        vpa = radians_to_degrees(atan2(vy,sd*vx))
        r = pp.ball_radius
        rb = self.received_ball
        t,bp,vm,w = rb.time, rb.position, rb.velocity, rb.spin
        # Measure time from paddle strike to return hit.
        t -= self.strike_time()
        vbm = norm(vm)
        fa = radians_to_degrees(atan2(vm[1],sd*vm[0]))
        th = pp.table_height
        h = bp[1] - th
        tlh = 0.5*pp.table_length
        d = -tlh + sd*bp[0]
        pl = self.paddle_length
        ib = self.incoming_ball
        bv = ib.velocity
        vb = norm(bv)
        ba = radians_to_degrees(atan2(bv[1],-sd*bv[0]))
        bs = ib.spin
        sp = ib.position
        bh = sp[1] - th
        bd = -tlh - sd*sp[0]
        op = self.outgoing_path
        bot, bposx, bposy, bovx, bovy, bow = (op[0][0] if op else (0,0,0,0,0,0))
        boa = radians_to_degrees(atan2(bovy,sd*bovx))
        thtext = '%.2f' % self.toss_height if self.serve else ''
        pp = self.physical_parameters
        fg = pp.ball_mass * pp.gravity
        from . import bounce
        fh = bounce.strike(bv, bs, vp, pa, r, pp.paddle_damping, pp.paddle_friction)[2]
        fd = bounce.drag_force(vm, pp) / fg if fg > 0 else 0
        fl = sd * bounce.lift_force(vm, w, pp) / fg if fg > 0 else 0

        if self.show_ranges and self.in_play:
            if self.param_ranges:
                (vbr, bar, tohr, bhr, bdr, bsr,
                 padr, psr, vfr, vsr, vpar, tmr) = self.param_ranges
            else:
                pr = self.param_range
                if self.serve:
                    tohmin, tohmax = pr(dh = 0.01)
                    tohr = '%5.2f  %5.2f' % (tohmin, tohmax)
                    vbr = bar = ''
                else:
                    tohr = ''
                    vbmin, vbmax = pr(dvb = normalize(bv, 0.2, (-sd,0)))
                    vbr = '%5.1f  %5.1f' % (vbmin,vbmax)
                    vbamin, vbamax = [radians_to_degrees(atan2(sin(a),-sd*cos(a)))
                                      for a in pr(dvba = -sd*pi/180.0)]
                    bar = '%6.1f  %5.1f' % (vbamin,vbamax)
                spmin, spmax = pr(dp = (0,0.01))
                bhmin, bhmax = spmin[1] - th, spmax[1] - th
                bhr = '%5.2f  %5.2f' % (bhmin, bhmax)
                spmin, spmax = pr(dp = (0.01,0))
                bdmin, bdmax = -tlh - sd*spmin[0], -tlh - sd*spmax[0]
                bdr = '%5.2f  %5.2f' % (bdmin, bdmax)
                wmin, wmax = pr(dw = sd*0.1/r)
                bsr = '%5.1f  %5.1f' % (sd*wmin*r,sd*wmax*r)
                pamin, pamax = pr(da = sd*pi/180.0)
                padmin, padmax = [side_angle(a,sd) for a in (pamin, pamax)]
                padr = '%5.1f  %5.1f' % (padmin,padmax)
                psmin, psmax = pr(dv = normalize(vp, 0.2, (sd,0)))
                psr = '%5.1f  %5.1f' % (psmin,psmax)
                vfmin, vfmax = pr(dv = (0.2*cpa,0.2*spa))
                vfr = '%5.1f  %5.1f' % (vfmin,vfmax)
                vsmin, vsmax = pr(dv = (-0.2*spa*sd,0.2*cpa*sd))
                vsr = '%5.1f  %5.1f' % (vsmin,vsmax)
                vpamin, vpamax = [radians_to_degrees(atan2(sin(a),sd*cos(a)))
                                  for a in pr(dva = sd*pi/180.0)]
                vpar = '%5.1f  %5.1f' % (vpamin,vpamax)
                tm = self.time_sensitivity(vp, pa, pl, bv)
                tmr = '' if tm is None else '%6.3f  %5.3f' % (-tm/2,tm/2)
                self.param_ranges = (vbr, bar, tohr, bhr, bdr, bsr,
                                     padr, psr, vfr, vsr, vpar, tmr)
        else:
            vbr = bar = tohr = bhr = bdr = bsr = padr = psr = vfr = vsr = vpar = tmr = ''

        f = self.fields
        pval = (('Incoming ball', (('toss', thtext, tohr),
                                   ('speed', '%.1f' % vb, vbr),
                                   ('slope', '%.1f' % ba, bar),
                                   ('spin', '%.1f' % (sd*bs*r), bsr),
                                   ('height', '%.2f' % bh, bhr),
                                   ('depth', '%.2f' % bd, bdr))),
                ('Hit', (('tilt', '%.1f' % pad, padr),
                         ('speed', '%.1f' % ps, psr),
                         ('flat', '%.1f' % vf, vfr),
                         ('slice', '%.1f' % vs, vsr),
                         ('swing', '%.1f' % vpa, vpar),
                         ('grip', '%.1f' % fh, ''),
                         ('time', '%.0f' % 0, tmr))),
                ('Outgoing ball', (('speed', '%.1f' % norm((bovx,bovy)), ''),
                                   ('slope', '%.1f' % boa, ''),
                                   ('spin', '%.1f' % (-sd*bow*r,), ''))),
                ('Return', (('time', '%.3f' % (t,), ''),
                            ('speed', '%.1f' % (vbm,), ''),
                            ('slope', '%.1f' % (fa,), ''),
                            ('spin', '%.1f' % (-sd*w*r,), ''),
                            ('height', '%.2f' % (h,), ''),
                            ('depth', '%.2f' % (d,), ''),
                            ('drag', '%.2f' % (fd,), ''),
                            ('lift', '%.2f' % (fl,), '')))

                )
        for h, plist in pval:
            for pname, val, rn in plist:
                p, r = f[(h, pname)]
                p.setText(val)
                r.setText(rn)

        if self.show_ranges:
            for p,r in f.values():
                r.show()
        else:
            for p,r in f.values():
                r.hide()

    def strike_time(self):

        return self.incoming_ball.time

    def receive_bounce_time(self):

        i = self.receive_curve_index()
        op = self.outgoing_path
        t = op[i][0][0] if i < len(op) else op[-1][-1][0]
        return t
        
    def receive_curve_index(self):

        ri = 2 if self.serve else 1
        return ri
        
    def time_sensitivity(self, vp, pa, pl, vb):

        # Time difference between when paddle edges would strike ball.
        det = -vp[0]*vb[1] + vb[0]*vp[1]
        if det == 0:
            return None
        tm = pl*((vp[0]-vb[0]) * cos(pa) + (vp[1]-vb[1])*sin(pa)) / det
        return abs(tm)

    def param_range(self, dh = 0, dp = (0,0), dvb = (0,0), dvba = 0, dw = 0,
                    da = 0, dv = (0,0), dva = 0,
                    steps = 100):

        side, serve, h = self.side, self.serve, self.toss_height
        ib = self.incoming_ball
        t0, sp, vb, w = ib.time, ib.position, ib.velocity, ib.spin
        a, v = self.paddle_angle, self.paddle_velocity
        vn = norm(v)
        va0 = atan2(v[1],v[0])
        vbn = norm(vb)
        vba0 = atan2(vb[1],vb[0])
        def inplay(s):
            if dva:
                va = va0 + s*dva
                vp = (vn*cos(va), vn*sin(va))
            else:
                vp = (v[0]+s*dv[0],v[1]+s*dv[1])
            spos = (sp[0]+s*dp[0], sp[1]+s*dp[1])
            if dvba:
                vba = vba0 + s*dvba
                vball = (vbn*cos(vba), vbn*sin(vba))
            else:
                vball = (vb[0]+s*dvb[0], vb[1]+s*dvb[1])
            return self.hit_flight(side, serve, h+s*dh, t0, spos, vball, w+s*dw,
                                   a+s*da, vp, path = False)
        for smin in range(0,-steps,-1):
            if not inplay(smin):
                break
        for smax in range(steps):
            if not inplay(smax):
                break
        if dh:
            return h+smin*dh, h+smax*dh
        elif dp != (0,0):
            return ((sp[0]+smin*dp[0],sp[1]+smin*dp[1]),
                    (sp[0]+smax*dp[0],sp[1]+smax*dp[1]))
        elif dvb != (0,0):
            dvbn = norm(dvb)
            vbdvb = (vb[0]*dvb[0]+vb[1]*dvb[1]) / dvbn
            return vbdvb+smin*dvbn, vbdvb+smax*dvbn
        elif dvba:
            return vba0+smin*dvba, vba0+smax*dvba
        elif dw:
            return w+smin*dw, w+smax*dw
        elif da:
            return a+smin*da, a+smax*da
        elif dv != (0,0):
            dvn = norm(dv)
            vdv = (v[0]*dv[0]+v[1]*dv[1]) / dvn
            return vdv+smin*dvn, vdv+smax*dvn
        elif dva:
            return va0+smin*dva, va0+smax*dva


    def flight_parameters(self):

        ib = self.incoming_ball
        ob = self.outgoing_ball
        p = {'side':self.side, 'serve':self.serve, 'strike':self.strike,
             'toss':self.toss_height,
             'ball position':ib.position,
             'ball velocity':ib.velocity,
             'ball spin':ib.spin,
             'ball velocity out':ob.velocity,
             'ball spin out':ob.spin,
             'paddle angle':self.paddle_angle,
             'paddle velocity':self.paddle_velocity,
             'time step':self.time_step}
        for n,v in self.physical_parameters.__dict__.items():
            if not n.startswith('_') and isinstance(v, (int,float)):
                p[n] = v
        return p

    def flight_changed(self, ignore_paddle_velocity = False):

        fparam = self.flight_parameters()
        prev_fparam = self.last_flight_params
        if ignore_paddle_velocity and prev_fparam:
            fparam = fparam.copy()
            prev_fparam = prev_fparam.copy()
            del fparam['paddle velocity'], prev_fparam['paddle velocity']
        return fparam != prev_fparam

    def calculate_flight(self):

        side, serve, strike = self.side, self.serve, self.strike
        ib = self.incoming_ball
        ob = self.outgoing_ball
        if strike:
            h = self.toss_height
            t0, sp, vb, w = ib.time, ib.position, ib.velocity, ib.spin
            a, v = self.paddle_angle, self.paddle_velocity
            self.in_play, curves = self.hit_flight(side, serve, h,
                                                   t0, sp, vb, w, a, v)
            if serve:
                c = curves[0]
                hs = c[-1]
                from .bounce import Ball_State
                self.incoming_ball = ib = Ball_State(time = hs[0],
                                                     position = tuple(hs[1:3]),
                                                     velocity = tuple(hs[3:5]),
                                                     spin = hs[5],
                                                     incoming_curve = c)
                curves = curves[1:]
            cout = curves[0][0]
            ob.velocity = tuple(cout[3:5])
            ob.spin = cout[5]
        else:
            pp, ts = self.physical_parameters, self.time_step
            t0, sp, v, w = ob.time, ob.position, ob.velocity, ob.spin
            from .bounce import flight
            self.in_play, curves = flight(t0, sp, v, w, side, serve, pp, ts)

        self.outgoing_path = curves
        self.last_flight_params = self.flight_parameters()

        # Place ball marker on 4th or 2nd arc.
        self.set_receive_position()

    def flight_path(self):

        fp = self.outgoing_path
        ib = self.incoming_ball
        ic = ib.incoming_curve
        if not ic is None:
            fp = [ic[:ib.curve_index+1]] + fp
        return fp
            
    def set_receive_position(self):

        pb = self.receive_curve_index()
        curves = self.outgoing_path
        if len(curves) > pb:
            c = curves[pb]
            if self.receive_deep:
                t = self.receive_deep_time(c)
            else:
                # Receive ball at top of bounce.
                t = max([(py,t) for t,px,py,vx,vy,w in c])[1]
                if t - self.strike_time() < 0.5:
                    t = self.receive_deep_time(c)
        else:
            c = curves[-1]
            t = c[-1][0]  # Not enough bounces (hit floor).
        self.set_receive_time(t, [c])

    def receive_deep_time(self, curve):

        # Receive ball at table level.
        pp = self.physical_parameters
        tlh = 0.5*pp.table_length
        if abs(curve[-1][1]) > tlh:
            th = pp.table_height
            for t,px,py,vx,vy,w in curve:
                if py < th and abs(px) > tlh:
                    break
        else:
            # Receive at top of bounce if next bounce is on table.
            t = max([(py,t) for t,px,py,vx,vy,w in curve])[1]
        return t

    def hit_flight(self, side, serve, h, t, sp, vb, w, a, v, path = True):

        curves = []
        pp = self.physical_parameters
        ts = self.time_step
        from .bounce import drop, hit
        if serve:
            f = drop(h, sp, side, serve, pp, ts)
            td, spx, spy, vbx, vby, w0 = f[-1]
            sp = (spx, spy)
            vb = (vbx, vby)
            if path:
                f[:,0] += t
                curves.append(f)
        else:
            td = 0
        maxpath = 10000 if path else 0
        r = hit(t+td, sp, vb, w, a, v, side, serve, pp, ts, maxpath = maxpath)
        if path:
            in_play, c = r
            curves.extend(c)
            return in_play, curves
        return r

    def set_receive_time(self, t0, path):

        bounce.set_ball_state_at_time(self.received_ball, t0, path)

    def draw_curves(self, painter, curves, scale = 1.0):

        points = []
        for c in curves:
            for t,px,py,vx,vy,w in c:
                p = QtCore.QPoint(*self.xy_to_pixel((px,py)))
                points.append(p)
        painter.setPen(Qt.black);
        painter.setBrush(Qt.black)
        painter.drawPolyline(*tuple(points))

    def paint_table(self, painter):

        painter.setPen(Qt.black)
        painter.setBrush(self.table_color)
        pp = self.physical_parameters
        s = self.scale
        w = pp.table_length
        h = pp.table_height
        t = self.table_thickness
        tc = self.xy_to_pixel((-w/2,h))
        table_top = QtCore.QRect(tc[0], tc[1], s*w, s*t)
        painter.drawRect(table_top)
        painter.setBrush(self.table_leg_color)
        ww,wh = self.graphics_window_size()
        lt = self.leg_thickness
        lp = self.leg_position
        l1c = self.xy_to_pixel((-lp*(w+lt)/2, h-t))
        l2c = self.xy_to_pixel((lp*(w+lt)/2, h-t))
        legs = [QtCore.QRectF(l1c[0], l1c[1], s*lt, s*h),
                QtCore.QRectF(l2c[0], l2c[1], s*lt, s*h)]
        painter.drawRects(legs)
        b = QtGui.QBrush(self.net_color, Qt.Dense4Pattern)
        painter.setBrush(b)
        hn = pp.net_height
        nt = self.net_thickness
        nc = self.xy_to_pixel((-nt/2, h+hn))
        net = QtCore.QRect(nc[0], nc[1], s*nt, s*hn)
        painter.drawRect(net)

    # Convert physical units to pixel painting position.
    def xy_to_pixel(self, p):

        x, y = p
        s = self.scale
        w, h = self.graphics_window_size()
        x0 = w/2 + self.xoffset
        return x0 + x*s, h - y*s

    def xy_to_pixel_matrix(self):

        s = self.scale
        w, h = self.graphics_window_size()
        x0 = w/2 + self.xoffset
        m = QtGui.QTransform(s, 0, 0, -s, x0, h)
        return m

    def pixel_to_xy(self, p):

        px, py = p
        w, h = self.graphics_window_size()
        px0 = w/2 + self.xoffset
        s = self.scale
        x,y = (px-px0)/s, (h-py)/s
        return x,y

    def graphics_window_size(self):

        g = self.graphics
        return g.width(), g.height()

    def click_side(self, pos):

        x, y = self.pixel_to_xy((pos.x(),pos.y()))
        side = 'right' if x >=0 else 'left'
        return side

    def paddle_rectangle(self, offset = (0,0)):

        x0,y0 = self.paddle_center(pixels = False)
        x0 += offset[0]
        y0 += offset[1]
        a = self.paddle_angle
        w = self.paddle_thickness
        h = self.paddle_length
        r = rotated_rectangle(x0,y0,w,h,a)
        rp = self.xy_to_pixel_matrix().map(r)
        return rp

    def paint_paddle(self, painter, offset = (0,0)):

        painter.setPen(Qt.black)
        painter.setBrush(self.paddle_color)
        painter.drawPolygon(self.paddle_rectangle(offset))

    def paddle_center(self, pixels = True):

        ib = self.incoming_ball
        sx,sy = ib.position
        t = self.paddle_thickness
        pp = self.physical_parameters
        r = pp.ball_radius
        d = t/2 + r
        a = self.paddle_angle
        c = (sx - d*cos(a), sy - d*sin(a))
        if pixels:
            return self.xy_to_pixel(c)
        return c

    def paddle_click(self, pos):

        r = self.paddle_rectangle()
        return r.containsPoint(pos, Qt.OddEvenFill)

    def paddle_center_click(self, pos):

        x,y = self.paddle_center()
        d = 0.25 * self.paddle_length * self.scale
        dx,dy = pos.x()-x, pos.y()-y
        return abs(dx) < d and abs(dy) < d

    def paint_ball(self, painter, center, spin = None):

        pp = self.physical_parameters
        rb = pp.ball_radius
        r = rb * self.scale
        x,y = self.xy_to_pixel(center)
        painter.setPen(Qt.black)
        painter.setBrush(self.ball_color)
        painter.drawEllipse(x-r, y-r, 2*r, 2*r)
        if spin and self.show_spin:
            angle = 2*pi * spin * rb / self.rotation_speed_scale
            w = self.spin_arrow_width * self.scale
            draw_circle_arrow(painter, (x,y), angle, r+3*w, w, self.spin_color)

    def top_of_toss(self):

        ib = self.incoming_ball
        sx,sy = ib.position
        t = self.toss_height
        return (sx, sy+t)

    def ball_click(self, cpos, bpos):

        return self.click_near_point(cpos, bpos, 1)

    def click_near_point(self, p, xy, rfactor):

        px, py = self.xy_to_pixel(xy)
        pp = self.physical_parameters
        d = rfactor * pp.ball_radius * self.scale
        dx,dy = p.x()-px, p.y()-py
        return dx*dx + dy*dy <= d*d

    def bounce_click(self, p):

        path = self.outgoing_path
        c = 2 if self.serve else 1
        if len(path) <= c:
            return False
        return self.click_near_point(p, path[c][0,1:3], 2)

    def over_net_click(self, p):

        path = self.outgoing_path
        c = 1 if self.serve else 0
        if len(path) <= c:
            return False
        y = bounce.height_at_zero(path[c][:,1:3])
        return self.click_near_point(p, (0,y), 2)

    def paint_paddle_velocity(self, painter):

        painter.setPen(Qt.black)
        painter.setBrush(self.velocity_arrow_color)
        painter.drawPolygon(self.paddle_velocity_arrow())

    def paddle_velocity_arrow(self, head_only = False):

        x, y = self.paddle_center(pixels = False)
        vx, vy = self.paddle_velocity
        v = sqrt(vx*vx + vy*vy)
        s = self.speed_scale
        pp = self.physical_parameters
        r = pp.ball_radius
        # Offset base of arrow from paddle center
        s0 = min(0.5*s, 4*r/v if v > 0 else 0.5*s)
        a = arrow((x+s0*vx,y+s0*vy), (x+s*vx,y+s*vy),
                  self.velocity_arrow_width, head_only = head_only)
        ap = self.xy_to_pixel_matrix().map(a)
        return ap

    def paddle_velocity_click(self, pos):

        a = self.paddle_velocity_arrow(head_only = True)
        c = a.containsPoint(pos, Qt.OddEvenFill)
        return c

    def under_table_click(self, pos):

        pp = self.physical_parameters
        tw, th, lp = pp.table_length, pp.table_height, self.leg_position
        w = lp*tw
        x0,y0 = self.xy_to_pixel((-w/2, 0))
        x1,y1 = self.xy_to_pixel((w/2, th))
        r = QtCore.QRect(x0, y0, x1-x0, y1-y0)
        return r.contains(pos)

    def calculate_vscan(self):

        side, serve, h = self.side, self.serve, self.toss_height
        ib = self.incoming_ball
        sp, vb, w = ib.position, ib.velocity, ib.spin
        pa = self.paddle_angle
        pp = self.physical_parameters
        ts = self.time_step
        from .bounce import drop, velocity_scan
        if serve:
            f = drop(h, sp, side, serve, pp, ts)
            t0, spx, spy, vbx, vby, w0 = f[-1]
            sp = (spx, spy)
            vb = (vbx, vby)
        vrange = self.vscan_range[side]
        vscan = velocity_scan(sp, vb, w, pa, side, serve, pp, vrange, ts)
        self.vscan = vscan
        self.vscan_image = None

    def paint_vscan(self, pt):

        if self.vscan_image is None:
            if self.vscan is None:
                self.calculate_vscan()
            s = self.vscan
            h,w = s.shape
            vcolors = [0xffffffff,] + [c.rgba() for c in self.vscan_colors]
            from numpy import array, uint32
            cmap = array(vcolors, uint32)
            rgba = cmap[s[::-1,:]]
            i = QtGui.QImage(rgba.data, w, h, QtGui.QImage.Format_RGB32)
            self.vscan_image = i
            i.save_rgba = rgba  # Keep rgba array from being freed.

        cx, cy = self.paddle_center(pixels = False)
        sc = self.speed_scale
        (vxmin, vxmax, vxs), (vymin, vymax, vys) = self.vscan_range[self.side]
        x0,y0 = self.xy_to_pixel((cx+sc*vxmin, cy+sc*vymax))
        x1,y1 = self.xy_to_pixel((cx+sc*vxmax, cy+sc*vymin))
        r = QtCore.QRectF(x0,y0,x1-x0,y1-y0)
        pt.drawImage(r, self.vscan_image)
        return

    def marker_click(self, p):
                
        r = 10
        r2 = r*r
        px, py = p.x(), p.y()
        cx, cy = self.paddle_center(pixels = False)
        s = self.speed_scale
        for a,vx,vy in self.markers:
            x,y = self.xy_to_pixel((cx + s*vx, cy + s*vy))
            dx, dy = px-x, py-y
            if dx*dx + dy*dy <= r2:
                self.paddle_angle = a
                self.paddle_velocity = (vx,vy)
                self.redraw()
                return

from PyQt5.QtWidgets import QWidget
class Ping_Pong_Graphics(QWidget):

    def __init__(self, parent):

        self.p = parent

        QWidget.__init__(self, parent.centralWidget())
        self.setFocusPolicy(Qt.ClickFocus)

        self.setPalette(QtGui.QPalette(parent.graphics_background))
        self.setAutoFillBackground(True)

    def resizeEvent(self, event):

        # When resize smaller may not get repaint event, but graphics
        # need to be redrawn since center of window changed.
        self.update()

    def mousePressEvent(self, event):

        self.p.mouse_press(event)
                
    def mouseMoveEvent(self, event):

        self.p.mouse_move(event)

    def mouseReleaseEvent(self, event):

        self.p.mouse_release(event)

    def wheelEvent(self, event):

        self.p.mouse_wheel(event)

    def paintEvent(self, event):

        self.p.paint(event)

def rotated_rectangle(center_x, center_y, width, height, angle):

    x0,y0,w,h,a = center_x, center_y, width, height, angle
    xy = ((w/2,h/2), (-w/2,h/2), (-w/2,-h/2), (w/2,-h/2))
    c,s = cos(a),sin(a)
    rsxy = [(c*x-s*y+x0, s*x+c*y+y0) for x,y in xy]
    points = [QtCore.QPointF(x,y) for x,y in rsxy]
    r = QtGui.QPolygonF(points)
    return r

def arrow(xy0, xy1, width, hf = 2, head_only = False):
    (x0,y0),(x1,y1) = xy0,xy1
    dx, dy = x1-x0, y1-y0
    length = sqrt(dx*dx+dy*dy)
    l,w = length, width
    xy = ((0,0.5*w),(l-hf*w,0.5*w),(l-hf*w,hf*w),(l,0),
          (l-hf*w,-hf*w),(l-hf*w,-0.5*w),(0,-0.5*w))
    if head_only:
        xy = xy[1:-1]
    c,s = normalize((dx,dy))
    rsxy = [(c*x-s*y+x0, s*x+c*y+y0) for x,y in xy]
    points = [QtCore.QPointF(x,y) for x,y in rsxy]
    a = QtGui.QPolygonF(points)
    return a

def draw_circle_arrow(painter, center, angle, radius, width, color,
                      astep = 0.1):

    r,w = radius, width
    ha = (2.0*w) / r
    if angle < 0:
        ha = -ha
    sa = angle - ha

    rgrow = 2*w/pi if abs(angle) > 2*pi else 0
        
    # Create arrow outline.
    pts = []
    div = int(ceil(abs(sa / astep)))
    # Inner half of arrow shaft.
    for s in range(div+1):
        a = float(sa*s)/div
        d = r - 0.5*w + abs(a)*rgrow
        pts.append((d*cos(a),d*sin(a)))
    # Add arrow head.
    d = r + abs(sa)*rgrow
    dt = r + abs(angle)*rgrow
    pts.extend([((d-1.5*w)*cos(sa), (d-1.5*w)*sin(sa)),
                (dt*cos(angle), dt*sin(angle)),
                ((d+1.5*w)*cos(sa), (d+1.5*w)*sin(sa))])
    # Outer half of arrow shaft.
    for s in range(div+1):
        a = float(sa*(div-s))/div
        d = r + 0.5*w + abs(a)*rgrow
        pts.append((d*cos(a),d*sin(a)))

    # Recenter arrow. Correct for screen y-axis pointing down.
    cx, cy = center
    pts = [(x+cx,cy-y) for x,y in pts]

    # Draw arrow
    points = [QtCore.QPointF(x,y) for x,y in pts]
    arrow = QtGui.QPolygonF(points)
    painter.setPen(Qt.black)
    painter.setBrush(color)
    painter.drawPolygon(arrow)

def side_angle(a, sd):
    c, s = cos(a), sin(a)
    ad = radians_to_degrees(atan2(s,sd*c))
    return ad

def radians_to_degrees(a):
    return a*180.0/pi

def norm(v):
    return sqrt(sum([x*x for x in v]))

def normalize(v, length = 1, v0 = (1,0)):
    n = norm(v)
    if n == 0:
        v = v0
        n = norm(v)
    u = (length*v[0]/n, length*v[1]/n)
    return u

def report_exception(type, value, traceback, parent):
    from traceback import format_exception
    msg = "".join(format_exception(type, value, traceback))
    import sys
    sys.stderr.write(msg + '\n')
    from PyQt5.QtWidgets import QMessageBox
    QMessageBox.warning(parent, 'Error', msg)

def start_gui(window_size = (1000,500), window_position = (100,100)):
    import sys
    from PyQt5.QtWidgets import QApplication
    app = QApplication(sys.argv)
    ppwin = Ping_Pong()
    sys.excepthook = lambda *a, p=ppwin: report_exception(*a, parent=p)
    x,y = window_position
    w,h = window_size
    ppwin.setGeometry(x, y, w, h)
    ppwin.show()
    sys.exit(app.exec_())

#start_gui()
